#include <sys/param.h>
#include <sys/conf.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/unistd.h>
#include <sys/kthread.h>
#include <sys/kernel.h>

#include <machine/bus.h>

#include "ptx.h"
#include "ptx_tuner.h"

#include "ptx_proc.h"

static int ring_pos = 0;
static int data_pos = 0;

static void
ptx_daemon(void* data);
static void
read_dmabuf(struct ptx_softc *scp, uint32_t *dataptr);
static void
reset_dma(struct ptx_softc *scp);

int
ptx_proc_start(struct ptx_softc *scp)
{
	int err = kthread_add(ptx_daemon, scp, NULL,
	    &scp->ptxdaemon, RFTHREAD, 0, "ptxdaemon");
	if (err) {
		device_printf(scp->device, "ptx_proc_start: kthread_add retured %d\n", err);
		return ENXIO;
	}

	scp->procrun = TRUE;
	return 0;
}

void
ptx_proc_stop(struct ptx_softc *scp)
{
	scp->procrun = FALSE;
}

static void
ptx_daemon(void* data)
{
	struct ptx_softc *scp = (struct ptx_softc *) data;
//	struct ptx_stream *s;

	uint32_t *dataptr;

	reset_dma(scp);

	while (scp->procrun == TRUE) {
		while (scp->procrun == TRUE) {

			dataptr = scp->databuf[ring_pos][data_pos].va;

			// ǡꡩ
			if (dataptr[(DMA_PAGE_SIZE / sizeof(uint32_t)) - 2] == 0) {
				break;
			}

			data_pos += 1;

			read_dmabuf(scp, dataptr);

			dataptr[(DMA_PAGE_SIZE / sizeof(uint32_t)) - 2] = 0;

			if (data_pos >= DMA_BUFFER_PAGE_COUNT) {
				data_pos = 0;
				ring_pos += 1;
				// DMA󥰤Ѥäϥ󥯥
				bus_space_write_4(scp->bt, scp->bh, 0x0, 0x00000020);
				if (ring_pos >= DMA_VIRTUAL_COUNTXSIZE) {
					ring_pos = 0;
				}
			}
		}
		ptx_pause("ptxdmn", MSTOTICK(1));
	}

	kthread_exit();
}

static void
read_dmabuf(struct ptx_softc *scp, uint32_t *dataptr)
{
	struct ptx_stream *s;
	int streamid;

	int lp;
	int packet_pos;

	uint32_t val;

	union mpacket {
		uint32_t val;
		MICRO_PACKET packet;
	} micro;

	for (lp = 0; lp < (DMA_PAGE_SIZE / sizeof(uint32_t)); lp++, dataptr++) {
		micro.val = *dataptr;
		streamid = ((micro.packet.head >> 5) & 0x07);
		//ͥ
		if (streamid > MAX_STREAM) {
			device_printf(scp->device, "DMA Channel Number Error(%d)\n", streamid);
			continue;
		}

		packet_pos = ((micro.packet.head >> 2) & 0x07);
		s = &scp->stream[streamid - 1];
		//  顼å
		if (micro.packet.head & MICROPACKET_ERROR) {
			device_printf(scp->device, "stream %d micropacket error\n", streamid);

			val = bus_space_read_4(scp->bt, scp->bh, 0x0);
			if (val & BIT_RAM_OVERFLOW) {
				s->overflow += 1;
			}
			if (val & BIT_INITIATOR_ERROR) {
				s->counetererr += 1;
			}
			if (val & BIT_INITIATOR_WARNING) {
				s->transerr += 1;
			}
			// Ƭ
			reset_dma(scp);
			ring_pos = data_pos = 0;
			break;
		}
		// ̤ѥͥϼΤƤ
		if (s->opened == FALSE) {
			continue;
		}

		// ƬǡХåե˻ĤäƤ
		if ((micro.packet.head & 0x02) &&
		    (s->chunk_filled % PACKET_SIZE != 0)) {
			int mod = s->chunk_filled % PACKET_SIZE;
			s->chunk_filled -= mod;
		}

		if (s->chunk_filled >= DATA_CHUNK_SIZE) {
			// ν񤭹

			mtx_lock(&s->lock);
			if ( ((s->wp + 1) % DATA_CHUNK_NUM) == s->rp ) {
				// оݥͥΥХåեդ줿
				cv_timedwait(&s->not_full, &s->lock, MSTOTICK(500));
			}
			if ( ((s->wp + 1) % DATA_CHUNK_NUM) == s->rp ) {
				mtx_unlock(&s->lock);
				s->drop += 1;
				continue;
			}

			s->wp++;
			if (s->wp >= DATA_CHUNK_NUM)
				s->wp = 0;
			cv_signal(&s->not_empty);
			mtx_unlock(&s->lock);

			s->chunk_filled = 0;
		}

		// ǡԡ
		KASSERT(s->chunk_filled >= 0, "negative chunk_filled?");
		KASSERT(s->chunk_filled < DATA_CHUNK_SIZE, "chunk_filled overflow");
		KASSERT(s->wp >= 0, "negative wp?");
		KASSERT(s->wp < DATA_CHUNK_NUM, "wp overflow");
		s->buf[s->wp * DATA_CHUNK_SIZE + s->chunk_filled] = micro.packet.data[2];
		s->chunk_filled++;
		s->buf[s->wp * DATA_CHUNK_SIZE + s->chunk_filled] = micro.packet.data[1];
		s->chunk_filled++;
		if (s->chunk_filled % PACKET_SIZE != 0) {
			s->buf[s->wp * DATA_CHUNK_SIZE + s->chunk_filled] = micro.packet.data[0];
			s->chunk_filled++;
		} else {
			// 189byteܤinformation
//printf("ptx: proc stream %d info=%02x\n", s->id, micro.packet.data[0]);
		}
	}
}

static void
reset_dma(struct ptx_softc *scp)
{
	int rpos;
	int dpos;
	uintptr_t addr;
	uint32_t addr32;
	uint32_t *dataptr;

	// ǡ
	for (rpos = 0; rpos < DMA_VIRTUAL_COUNTXSIZE; ++rpos) {
		for (dpos = 0; dpos < DMA_BUFFER_PAGE_COUNT; ++dpos) {
			dataptr = scp->databuf[rpos][dpos].va;
			dataptr[(DMA_PAGE_SIZE / sizeof(uint32_t)) - 2] = 0;
		}
	}

	// ž󥿤ꥻå
	bus_space_write_4(scp->bt, scp->bh, 0x00, 0x00000010);
	// ž󥿤򥤥󥯥
	for (rpos = 0; rpos < DMA_VIRTUAL_COUNTXSIZE; ++rpos) {
		bus_space_write_4(scp->bt, scp->bh, 0x00, 0x00000020);
	}

	addr = (uintptr_t) scp->ringbuf[0].pa;
	addr >>= 12;
	addr32 = addr & 0x7fffffff;
	// DMAХåե
	bus_space_write_4(scp->bt, scp->bh, DMA_ADDR, addr32);
	// DMA
	bus_space_write_4(scp->bt, scp->bh, 0x00, 0x0c000040);
}
